### Server side module

import pygame

import things
import server

max_pursuit_path = 20

def step():
    map = server.map
    for thing in map.things.values():
        if isinstance(thing,things.Monster):
            thing.maybe_fight()
            new_direction = thing.maybe_move()
            if new_direction:
                new_position = (thing.position[0] + new_direction[0],
                                thing.position[1] + new_direction[1])
                try:
                    thing2 = server.map.positions[new_position]
                    if isinstance(thing2, things.Monster):
                        if thing.debug or thing2.debug:
                            print "%s) Rude! %s" % (thing2.id, thing)
                        thing2.rude_monster = thing, thing.position
                except KeyError: pass
                if map.move(thing, new_direction[0],new_direction[1]):
                    thing.rude_monster = None
    map.maybe_generate_monsters()

# represents info about paths to a single destination point
class ShortestPaths:
    # @param destination  a pygame.Rect defining an inclusive
    #                     destination rectangle to navigate toward
    def __init__(self, destination):
        # breadth-first search
        self.distances = []
        for col in range(len(server.map.grid)):
            self.distances.append([])
            for row in range(len(server.map.grid[0])):
                self.distances[col].append(None)
        # nodes at a known distance, ordered by distance
        self.nodes = []
        
        for x in range(destination.left, destination.right + 1):
            for y in range(destination.top, destination.bottom + 1):
                self.distances[x][y] = 0
                self.nodes.append((x, y))
                
        # first node we haven't looked for neighbors around
        self.tail = 0

    def __str__(self):
        result = []
        for x, y in self.nodes:
            distance = self.distances[x][y]
            result.append('(%d, %d) : %d' % (x, y, distance))
        return ', '.join(result)

    # @param source        the source point to find the distance from
    #                        if None, generates all acceptable paths
    #                        to destination
    # @param max_distance  don't generate paths longer than max_distance
    #                        if None, generate paths of any length
    # @return              the distance from the source to destination
    #                        or None if no acceptable path exists
    #                        or no source specified
    def distance(self, source = None, max_distance = None):
        known_distance = self.distances[source[0]][source[1]]
        if known_distance is not None:
            if max_distance is None or known_distance <= max_distance:
                return known_distance
            else:
                return None

        result = None

        while self.tail < len(self.nodes):
            x, y = self.nodes[self.tail]
            new_distance = self.distances[x][y] + 1
            if max_distance is not None and new_distance > max_distance:
                return None
            
            for delta_x in (-1, 0, 1):
                for delta_y in (-1, 0, 1):
                    new_x = x + delta_x
                    new_y = y + delta_y
                    if (server.map.hopefully_movable_space(new_x, new_y) and
                        self.distances[new_x][new_y] is None):
                        self.distances[new_x][new_y] = new_distance
                        self.nodes.append((new_x, new_y))
                        if (new_x, new_y) == source: result = new_distance
                        # keep going, finish this node
                        
            self.tail += 1

            if result is not None: return result
            
        return None

    # @params same as for distance
    # @return a pair:
    #           the distance from the source to the destination,
    #           the next point on the path
    #         or 0, (None, None) if the source is the destination
    #         or None, (None, None) if no acceptable path exists
    def search(self, source, max_distance = None):
        distance = self.distance(source, max_distance)
        if not distance: return distance, (None, None)
        
        result = None
        
        x, y = source
        for delta_x in (-1, 0, 1):
            for delta_y in (-1, 0, 1):
                new_x = x + delta_x
                new_y = y + delta_y
                if (server.map.hopefully_movable_space(new_x, new_y) and
                    self.distances[new_x][new_y] is not None and
                    self.distances[new_x][new_y] < distance):
                    if not (delta_x and delta_y): #cardinal direction
                        return distance, (new_x, new_y)
                    else:
                        # equally short path, but prefer non-diagonal
                        result = (new_x, new_y)

        if result:
            return distance, result
        else:
            # shouldn't happen
            return None, (None, None)

    # @params same as for distance
    # @return a list of points along the shortest path
    #         from the source to the destination
    #         or None, if no acceptable path exists
    def path(self, source, max_distance = None):
        result = []
        while source != (None, None):
            result.append(source)
            distance, source = self.search(source, max_distance)
            if distance is None: return None
        return result
